/*
 SNEeSe 65c816 CPU emulation core
 Originally written by Savoury SnaX (Not quite sure if I like AT&T)
 Maintained/rewritten by Charles Bilyue'

 Compile under DJGPP (GNU-AS)
  NOTE - This must have .S (capital) or the # stuff wont be pre-processed!!!

 This file contains:
  65c816 addressing mode macros

 Note: As of 3-99, be very careful when modifying this code, and
 the CPU loops! The CPU loops set eax to 0, except al, and ebx to
 PB:PC+1 (start of immediate data), and esi to address of data at PB:PC,
 and the new address mode emulation expects this!

 Final address in ebx, when there is one!

*/

/*
 Immediate (read-only)
  Read 8-bit and 16-bit
    ORA-09  AND-29  EOR-49  ADC-69  BIT-89  LDY-A0  LDX-A2  LDA-A9
    CPY-C0  CMP-C9  CPX-E0  SBC-E9
  Read 8-bit
    REP-C2  SEP-E2
*/
.macro  ADDR_Immediate
.endm

.macro  READ8_Immediate
    incl %ebx
    movb 1(%esi),%al
    movw %bx,_PC
.endm

.macro  READ16_Immediate
    movb 1(%esi),%al
    addl $2,%ebx
    movb 2(%esi),%ah
    movw %bx,_PC
.endm

/*
 Absolute Indexed Indirect (corrupts eax)
 Returns new PCH in %al
  Address = (PB:X+a16)
  Address set to PC
    JMP-7C  JSR-FC
*/
.macro  ADDR_Absolute_Indexed_Indirect
    movb 1(%esi),%bl
    movb 2(%esi),%bh
    addw _X,%bx
    GET_BYTE_PB
    incw %bx
    movb %al,_PC
    GET_BYTE_PB
    movb %al,_PC+1
.endm

/*
 Absolute
  Address = DB:a16
  Read 8-bit and 16-bit
    ORA-0D  BIT-2C  AND-2D  EOR-4D  ADC-6D  LDY-AC  LDA-AD  LDX-AE
    CPY-CC  CMP-CD  CPX-EC  SBC-ED
  Write 8-bit and 16-bit
    STZ-9C  STY-8C  STA-8D  STX-8E
  RMW 8-bit and 16-bit
    TSB-0C  ASL-0E  TRB-1C  ROL-2E  LSR-4E  ROR-6E  DEC-CE  INC-EE
*/
.macro  ADDR_Absolute
    addl $2,%ebx
    movw %bx,_PC
    movl _DB_Shifted,%ebx
    movb 1(%esi),%bl
    movb 2(%esi),%bh
.endm

.macro  READ8_Absolute
    ADDR_Absolute
    GET_BYTE
.endm

.macro  READ16_Absolute Reverse=0
    ADDR_Absolute
    GET_WORD Reverse=\Reverse
.endm

/*
 Absolute Long (corrupts eax)
  Address = a24
  Read 8-bit and 16-bit
    ORA-0F  AND-2F  EOR-4F  ADC-6F  LDA-AF  CMP-CF  SBC-EF
  Write 8-bit and 16-bit
    STA-8F
*/
.macro  ADDR_Absolute_Long
    movb 3(%esi),%al
    addl $3,%ebx
    shll $16,%eax
    movw %bx,_PC
    movb 1(%esi),%al
    movb 2(%esi),%ah
    movl %eax,%ebx
.endm

.macro  READ8_Absolute_Long
    ADDR_Absolute_Long
    GET_BYTE
.endm

.macro  READ16_Absolute_Long Reverse=0
    ADDR_Absolute_Long
    GET_WORD Reverse=\Reverse
.endm

/*
 Direct (corrupts eax)
  Read 8-bit and 16-bit
    ORA-05  BIT-24  AND-25  EOR-45  ADC-65  LDY-A4  LDA-A5  LDX-A6
    CPY-C4  CMP-C5  CPX-E4  SBC-E5
  Write 8-bit and 16-bit
    STZ-64  STY-84  STA-85  STX-86
  RMW 8-bit and 16-bit
    TSB-04  ASL-06  TRB-14  ROL-26  LSR-46  ROR-66  DEC-C6  INC-E6
*/
.macro  ADDR_Direct
    incl %ebx
    movb 1(%esi),%al
    movw %bx,_PC
    addw _D,%ax
    movl %eax,%ebx
.endm

.macro  READ8_Direct
    ADDR_Direct
    GET_BYTE_00
.endm

.macro  READ16_Direct Reverse=0
    ADDR_Direct
    GET_WORD_00 Reverse=\Reverse
.endm

/*
 Direct Indirect Indexed (corrupts eax)
  Read 8-bit and 16-bit
    ORA-11  AND-31  EOR-51  ADC-71  LDA-B1  CMP-D1  SBC-F1
  Write 8-bit and 16-bit
    STA-91
*/
.macro  ADDR_Direct_Indirect_Indexed
    READ16_Direct Reverse=1
    movl _DB_Shifted,%ebx
    movb %al,%bh
    movb %ah,%bl
    addl _Y,%ebx
    andl $0x00FFFFFF,%ebx   // Ensure 24-bit address
.endm

.macro  READ8_Direct_Indirect_Indexed
    ADDR_Direct_Indirect_Indexed
    GET_BYTE
.endm

.macro  READ16_Direct_Indirect_Indexed Reverse=0
    ADDR_Direct_Indirect_Indexed
    GET_WORD Reverse=\Reverse
.endm

/*
 Direct Indirect Indexed Long (corrupts eax)
  Read 8-bit and 16-bit
    ORA-17  AND-37  EOR-57  ADC-77  LDA-B7  CMP-D7  SBC-F7
  Write 8-bit and 16-bit
    STA-97
*/
.macro  ADDR_Direct_Indirect_Indexed_Long
    READ16_Direct Reverse=1
    bswapl %eax
    incw %bx
    GET_BYTE_00
    rorl $16,%eax
    movl _Y,%ebx
    addl %eax,%ebx
    andl $0x00FFFFFF,%ebx   // Ensure 24-bit address
.endm

.macro  READ8_Direct_Indirect_Indexed_Long
    ADDR_Direct_Indirect_Indexed_Long
    GET_BYTE
.endm

.macro  READ16_Direct_Indirect_Indexed_Long Reverse=0
    ADDR_Direct_Indirect_Indexed_Long
    GET_WORD Reverse=\Reverse
.endm

/*
 Direct index X (corrupts eax)
  Read 8-bit and 16-bit
    ORA-15  BIT-34  AND-35  EOR-55  ADC-75  LDY-B4  LDA-B5  CMP-D5
    SBC-F5
  Write 8-bit and 16-bit
    STZ-74  STY-94  STA-95
  RMW 8-bit and 16-bit
    ASL-16  ROL-36  LSR-56  ROR-76  DEC-D6  INC-F6
*/
.macro  ADDR_Direct_Index_X
    movb 1(%esi),%al
    incl %ebx
    movw %bx,_PC
    addw _D,%ax
    addw _X,%ax
    movl %eax,%ebx
.endm

.macro  READ8_Direct_Index_X
    ADDR_Direct_Index_X
    GET_BYTE_00
.endm

.macro  READ16_Direct_Index_X Reverse=0
    ADDR_Direct_Index_X
    GET_WORD_00 Reverse=\Reverse
.endm

/*
 Direct index Y (corrupts eax)
  Read 8-bit and 16-bit
    LDX-B6
  Write 8-bit and 16-bit
    STX-96
*/
.macro  ADDR_Direct_Index_Y
    movb 1(%esi),%al
    incl %ebx
    movw %bx,_PC
    addw _D,%ax
    addw _Y,%ax
    movl %eax,%ebx
.endm

.macro  READ8_Direct_Index_Y
    ADDR_Direct_Index_Y
    GET_BYTE_00
.endm

.macro  READ16_Direct_Index_Y Reverse=0
    ADDR_Direct_Index_Y
    GET_WORD_00 Reverse=\Reverse
.endm

/*
 Direct Indexed Indirect (corrupts eax)
  Read 8-bit and 16-bit
    ORA-01  AND-21  EOR-41  ADC-61  LDA-A1  CMP-C1  SBC-E1
  Write 8-bit and 16-bit
    STA-81
*/
.macro  ADDR_Direct_Indexed_Indirect
    READ16_Direct_Index_X Reverse=1
    movl _DB_Shifted,%ebx
    movb %al,%bh
    movb %ah,%bl
.endm

.macro  READ8_Direct_Indexed_Indirect
    ADDR_Direct_Indexed_Indirect
    GET_BYTE
.endm

.macro  READ16_Direct_Indexed_Indirect Reverse=0
    ADDR_Direct_Indexed_Indirect
    GET_WORD Reverse=\Reverse
.endm

/*
 Absolute Index X (corrupts eax)
  Read 8-bit and 16-bit
    ORA-1D  BIT-3C  AND-3D  EOR-5D  ADC-7D  LDY-BC  LDA-BD  CMP-DD
    SBC-FD
  Write 8-bit and 16-bit
    STZ-9E  STA-9D
  RMW 8-bit and 16-bit
    ASL-1E  ROL-3E  LSR-5E  ROR-7E  DEC-DE  INC-FE
*/
.macro  ADDR_Absolute_Index_X
    addl $2,%ebx
    movw %bx,_PC
    movl _DB_Shifted,%ebx
    movb 1(%esi),%bl
    movb 2(%esi),%bh
    addl _X,%ebx
    andl $0x00FFFFFF,%ebx   // Ensure 24-bit address
.endm

.macro  READ8_Absolute_Index_X
    ADDR_Absolute_Index_X
    GET_BYTE
.endm

.macro  READ16_Absolute_Index_X Reverse=0
    ADDR_Absolute_Index_X
    GET_WORD Reverse=\Reverse
.endm

/*
 Absolute Index Y (corrupts eax)
  Read 8-bit and 16-bit
    ORA-19  AND-39  EOR-59  ADC-79  LDA-B9  LDX-BE  CMP-D9  SBC-F9
  Write 8-bit and 16-bit
    STA-99
*/
.macro  ADDR_Absolute_Index_Y
    addl $2,%ebx
    movw %bx,_PC
    movl _DB_Shifted,%ebx
    movb 1(%esi),%ah
    movb %ah,%bl
    movb 2(%esi),%ah
    movb %ah,%bh
    addl _Y,%ebx
    andl $0x00FFFFFF,%ebx   // Ensure 24-bit address
.endm

.macro  READ8_Absolute_Index_Y
    ADDR_Absolute_Index_Y
    GET_BYTE
.endm

.macro  READ16_Absolute_Index_Y Reverse=0
    ADDR_Absolute_Index_Y
    GET_WORD Reverse=\Reverse
.endm

/*
 Absolute Long Index X (corrupts eax)
  Read 8-bit and 16-bit
    ORA-1F  AND-3F  EOR-5F  ADC-7F  LDA-BF  CMP-DF  SBC-FF
  Write 8-bit and 16-bit
    STA-9F
*/
.macro  ADDR_Absolute_Long_Index_X
    movb 3(%esi),%al
    addl $3,%ebx
    shll $16,%eax
    movw %bx,_PC
    movb 1(%esi),%al
    movl _X,%ebx
    movb 2(%esi),%ah
    addl %eax,%ebx
    andl $0x00FFFFFF,%ebx   // Ensure 24-bit address
.endm

.macro  READ8_Absolute_Long_Index_X
    ADDR_Absolute_Long_Index_X
    GET_BYTE
.endm

.macro  READ16_Absolute_Long_Index_X Reverse=0
    ADDR_Absolute_Long_Index_X
    GET_WORD Reverse=\Reverse
.endm

/*
 Direct Indirect (corrupts eax)
  Read 8-bit and 16-bit
    ORA-12  AND-32  EOR-52  ADC-72  LDA-B2  CMP-D2  SBC-F2
  Write 8-bit and 16-bit
    STA-92
*/
.macro  ADDR_Direct_Indirect
    READ16_Direct Reverse=1
    movl _DB_Shifted,%ebx
    movb %al,%bh
    movb %ah,%bl
.endm

.macro  READ8_Direct_Indirect
    ADDR_Direct_Indirect
    GET_BYTE
.endm

.macro  READ16_Direct_Indirect Reverse=0
    ADDR_Direct_Indirect
    GET_WORD Reverse=\Reverse
.endm

/*
 Direct Indirect Long (corrupts eax)
  Read 8-bit and 16-bit
    ORA-07  AND-27  EOR-47  ADC-67  LDA-A7  CMP-C7  SBC-E7
  Write 8-bit and 16-bit
    STA-87
*/
.macro  ADDR_Direct_Indirect_Long
    READ16_Direct Reverse=1
    bswapl %eax
    incw %bx
    GET_BYTE_00
    rorl $16,%eax
    movl %eax,%ebx
.endm

.macro  READ8_Direct_Indirect_Long
    ADDR_Direct_Indirect_Long
    GET_BYTE
.endm

.macro  READ16_Direct_Indirect_Long Reverse=0
    ADDR_Direct_Indirect_Long
    GET_WORD Reverse=\Reverse
.endm

/*
 Stack Relative (corrupts eax)
  Read 8-bit and 16-bit
    ORA-03  AND-23  EOR-43  ADC-63  LDA-A3  CMP-C3  SBC-E3
  Write 8-bit and 16-bit
    STA-83
*/
.macro  ADDR_Stack_Relative
    incl %ebx
    movw %bx,_PC
    movb 1(%esi),%al
    movl _S,%ebx    // Add stack pointer to make address
    addb %al,%bl
    adcb %ah,%bh
.endm

.macro  READ8_Stack_Relative
    ADDR_Stack_Relative
    GET_BYTE_00
.endm

.macro  READ16_Stack_Relative Reverse=0
    ADDR_Stack_Relative
    GET_WORD_00 Reverse=\Reverse
.endm

/*
 Stack Relative Indirect Indexed (corrupts eax)
  Read 8-bit and 16-bit
    ORA-13  AND-33  EOR-53  ADC-73  LDA-B3  CMP-D3  SBC-F3
  Write 8-bit and 16-bit
    STA-93
*/
.macro  ADDR_Stack_Relative_Indirect_Indexed
    ADDR_Stack_Relative
    movl _DB_Shifted,%eax
    GET_WORD
    movl _Y,%ebx
    addl %eax,%ebx
    andl $0x00FFFFFF,%ebx   // Ensure 24-bit address
.endm

.macro  READ8_Stack_Relative_Indirect_Indexed
    ADDR_Stack_Relative_Indirect_Indexed
    GET_BYTE
.endm

.macro  READ16_Stack_Relative_Indirect_Indexed Reverse=0
    ADDR_Stack_Relative_Indirect_Indexed
    GET_WORD Reverse=\Reverse
.endm

/*
 Absolute (corrupts eax)
  Address = a16
  Address set to PC
    JSR-20  JMP-4C
*/
.macro  ADDR_Absolute_JMP
    movb 1(%esi),%al
    incl %ebx
    movb %al,_PC
    movb 2(%esi),%al
    movb %al,_PC+1
.endm
