/*

 SNEeSe v0.37

 main.cc    - This is the base file, it contains the calls to set up the
               emulation and a few other bits.
  wrapaleg.h - Wrapper to include Allegro with alleg_*_unused #defines.
  types.h    - Some GUI color #defines, and screen mode structure.
  romload.h  - Structure and data declarations used by the ROM loader.
  scrmode.h  - Declarations for special mode switch functions.
  cpu.h      - 65c816 emulation declarations.
  spc.h      - SPC700 emulation declarations.
  sound.h    - SPC700 internal DSP emulation declarations.
  helper.h   - Lots of externally defined variable declarations.
  mem.h      - Declarations from memory mapper code.
  guicore.h  - GUI interface core declarations.
  emugui.h   - Emulator-specific GUI declarations.
  apuskip.h  - APU skipper declarations.

 scrmode.S  - Handlers to set up special display modes

 romload.cc - ROM loading routine, decodes information such as memory model
               from the ROM image.
  romload.h  - Structure and data declarations used by the ROM loader.

 CPUmain.S  - 65c816 CPU reset/loop code.
  CPUmacro.S - 65c816 CPU instruction emulation macros.
  CPUaddr.S  - 65c816 CPU address mode emulation macros.
  CPUx.S     - 65c816 CPU opcode emulation code.
                CPUx.S contains opcodes (x0-xF)
  CPU.i      - 65c816 CPU aliases, constants, and flag and stack macros.
  mem.i      - Memory mapper read/write macros.

 SPCmain.S  - SPC700 CPU reset/loop code.
  SPCx.S     - SPC700 CPU opcode emulation code.
                SPCx.S contains opcodes (x0-xF)
  SPCmacro.S - SPC700 CPU instruction emulation macros.
  SPCaddr.S  - SPC700 CPU address mode emulation macros.

 helper.c   - Contains variable and function definitions used by both ASM
               and C++, defined as C for easier referencing.... NB I had to
               do this for vars cos, debugger was playing up!?

 sound.c    - SPC700 internal DSP emulation/sound generation routines.

 mem.S      - LoROM-specific mapping functions/macros/tables.
  memmap.i   - Memory mapping macros.
 memhi.S    - HiROM-specific mapping functions/macros/tables.
  memmap.i   - Memory mapping macros.
 memblank.S - Dynamic memory map and non-specific memory map routines.
  memmap.i   - Memory mapping macros.
 PPU.S      - Hardware port (21xx/40xx/42xx/43xx) handlers.
  key.i      - Keyboard checking macros.
 DMA.S      - (H)DMA emulation code.
  DMA.i      - (H)DMA related macros.

 screenL.S  - Base line-by-line screen renderers.
  irq.S      - Screen render timing/IRQ/NMI code.
   scankeys.h - A list of keyboard scan codes.
   key.i      - Keyboard checking macros.
   cpu.i      - CPU flag and stack macros.
   DMA.i      - (H)DMA related macros.

 tilesL.S   - Tile renderers.
 spriteL.S  - Sprite renderers.

 guicore.cc - Core code of GUI interface used for the emulator
  font4.h    - A 6x8 font used by the GUI.

 emugui.cc  - Emulator-specific GUI code for configuring the emulator,
               loading and resetting the emulation, etc.

*/

#define RELEASE_DATE "15 October 1999"
#define SNEESE_VERSION "0.37"

#include <stdio.h>
#include <conio.h>
#include <stdlib.h>
#include <string.h>
#include <dos.h>
#include <fstream.h>
#include <unistd.h>

#include "wrapaleg.h"   // Wrapper header for Allegro

#include "types.h"
#include "romload.h"
#include "scrmode.h"
#include "cpu.h"
#include "spc.h"
#include "sound.h"
#include "helper.h"
#include "mem.h"
#include "guicore.h"
#include "emugui.h"
#include "apuskip.h"

#include <crt0.h>

void   __crt0_load_environment_file(char *_app_name){}
char **__crt0_glob_function(char *_arg){ return 0; }

int _crt0_startup_flags = /*_CRT0_FLAG_NO_LFN |*/ _CRT0_FLAG_FILL_DEADBEEF;
// GUI file selector doesn't support LFNs yet - temp fix

#ifdef DEBUG
volatile unsigned long Timer_Counter_Profile = 0,
 CPU_Profile_Last = 0, CPU_Profile_Test = 0,
 SPC_Profile_Last = 0, SPC_Profile_Test = 0,
 GFX_Profile_Last = 0, GFX_Profile_Test = 0;
void Timer_Handler_Profile(...){ Timer_Counter_Profile++; }
END_OF_FUNCTION(Timer_Handler_Profile);

volatile unsigned long Timer_Counter_FPS=0;
void Timer_Handler_FPS(...){ Timer_Counter_FPS++; }
END_OF_FUNCTION(Timer_Handler_FPS);
#endif

extern "C" volatile unsigned FPSTicks;
volatile unsigned Timer_Counter_Throttle asm("Throttle");
void Timer_Handler_Throttle(...){
 Timer_Counter_Throttle++; FPSTicks++;
}
END_OF_FUNCTION(Timer_Handler_Throttle);

void LoadConfig(){
 PC_SCREEN_MODE = get_config_int(NULL, "screenmode", 0);
 if (PC_SCREEN_MODE > 7) PC_SCREEN_MODE = 0;

 // For backwards compatibility with old versions/config files
 int oldframeskip = get_config_int(NULL, "frameskip", 4);
 FRAME_SKIP_MAX = get_config_int(NULL, "maxframeskip", oldframeskip);
 if (FRAME_SKIP_MAX > 60) FRAME_SKIP_MAX = 60;
 if (FRAME_SKIP_MAX > 30 && FRAME_SKIP_MAX < 60) FRAME_SKIP_MAX = 30;
 if (FRAME_SKIP_MAX > 20 && FRAME_SKIP_MAX < 30) FRAME_SKIP_MAX = 20;
 if (FRAME_SKIP_MAX > 15 && FRAME_SKIP_MAX < 20) FRAME_SKIP_MAX = 15;
 if (FRAME_SKIP_MAX > 10 && FRAME_SKIP_MAX < 15) FRAME_SKIP_MAX = 10;
 if (FRAME_SKIP_MAX < 1) FRAME_SKIP_MAX = 1;
 FRAME_SKIP_MIN = get_config_int(NULL, "minframeskip", 0);
 if (FRAME_SKIP_MIN > 60) FRAME_SKIP_MIN = 60;
 if (FRAME_SKIP_MIN > 30 && FRAME_SKIP_MIN < 60) FRAME_SKIP_MIN = 30;
 if (FRAME_SKIP_MIN > 20 && FRAME_SKIP_MIN < 30) FRAME_SKIP_MIN = 20;
 if (FRAME_SKIP_MIN > 15 && FRAME_SKIP_MIN < 20) FRAME_SKIP_MIN = 15;
 if (FRAME_SKIP_MIN > 10 && FRAME_SKIP_MIN < 15) FRAME_SKIP_MIN = 10;
 if (FRAME_SKIP_MIN > FRAME_SKIP_MAX) FRAME_SKIP_MIN = FRAME_SKIP_MAX;
 JOYSTICK_ENABLED = get_config_int(NULL, "joystick", 0);
 if (JOYSTICK_ENABLED > 2 || (JOYSTICK_ENABLED == 2 && !MOUSE_PLUGGED_IN))
  JOYSTICK_ENABLED = 0;
 JOYSTICK_ENABLED2 = get_config_int(NULL, "joystick2", 0);
 if (JOYSTICK_ENABLED2 > 2 || (JOYSTICK_ENABLED2 == 2 && !MOUSE_PLUGGED_IN))
  JOYSTICK_ENABLED2 = 0;
 SPC_ENABLED = get_config_int(NULL, "emulatespc", 0);
 for (int a = 0; a < 4; a++) SaveExtension[a] = '\0';
 strncpy(SaveExtension, get_config_string(NULL, "saveextension", "RAM"), 3);
 SOUND_ENABLED = get_config_int(NULL, "outputsound", 0);
 if (SOUND_ENABLED > 2) SOUND_ENABLED = 0;

 char *keymapbuf = (char *) malloc(41);
 char default_keymap[] = "72 80 75 77 18 32 17 31 16 30 26 27";

 strncpy(keymapbuf, get_config_string(NULL, "snes_pad_1_keys", default_keymap), 40);
 if (
  sscanf(keymapbuf, "%X%X%X%X%X%X%X%X%X%X%X%X",
   &KEY_UP_1, &KEY_DOWN_1, &KEY_LEFT_1, &KEY_RIGHT_1,
   &KEY_A_1, &KEY_B_1, &KEY_X_1, &KEY_Y_1, &KEY_L_1, &KEY_R_1,
   &KEY_SELECT_1, &KEY_START_1) < 12){
  sscanf(default_keymap, "%X%X%X%X%X%X%X%X%X%X%X%X",
   &KEY_UP_1, &KEY_DOWN_1, &KEY_LEFT_1, &KEY_RIGHT_1,
   &KEY_A_1, &KEY_B_1, &KEY_X_1, &KEY_Y_1, &KEY_L_1, &KEY_R_1,
   &KEY_SELECT_1, &KEY_START_1);
 }

 strncpy(keymapbuf, get_config_string(NULL, "snes_pad_2_keys", default_keymap), 40);
 if (
  sscanf(keymapbuf, "%X%X%X%X%X%X%X%X%X%X%X%X",
   &KEY_UP_2, &KEY_DOWN_2, &KEY_LEFT_2, &KEY_RIGHT_2,
   &KEY_A_2, &KEY_B_2, &KEY_X_2, &KEY_Y_2, &KEY_L_2, &KEY_R_2,
   &KEY_SELECT_2, &KEY_START_2) < 12){
  sscanf(default_keymap, "%X%X%X%X%X%X%X%X%X%X%X%X",
   &KEY_UP_2, &KEY_DOWN_2, &KEY_LEFT_2, &KEY_RIGHT_2,
   &KEY_A_2, &KEY_B_2, &KEY_X_2, &KEY_Y_2, &KEY_L_2, &KEY_R_2,
   &KEY_SELECT_2, &KEY_START_2);
 }
}

void SaveConfig(){
 set_config_string(NULL, "saveextension", SaveExtension);
 set_config_int(NULL, "screenmode", PC_SCREEN_MODE);

 set_config_int(NULL, "minframeskip", FRAME_SKIP_MIN);
 set_config_int(NULL, "maxframeskip", FRAME_SKIP_MAX);
 // For backwards compatibility with old versions/config files
 set_config_int(NULL, "frameskip", FRAME_SKIP_MAX);

 set_config_int(NULL, "joystick", JOYSTICK_ENABLED);
 set_config_int(NULL, "joystick2", JOYSTICK_ENABLED2);
 set_config_int(NULL, "emulatespc", SPC_ENABLED);
 set_config_int(NULL, "outputsound", SOUND_ENABLED);

 char *keymapbuf = (char *) malloc(41);
 sprintf(keymapbuf,"%X %X %X %X %X %X %X %X %X %X %X %X",
  KEY_UP_1, KEY_DOWN_1, KEY_LEFT_1, KEY_RIGHT_1,
  KEY_A_1, KEY_B_1, KEY_X_1, KEY_Y_1, KEY_L_1, KEY_R_1,
  KEY_SELECT_1, KEY_START_1);
 set_config_string(NULL, "snes_pad_1_keys", keymapbuf);
 sprintf(keymapbuf, "%X %X %X %X %X %X %X %X %X %X %X %X",
  KEY_UP_2, KEY_DOWN_2, KEY_LEFT_2, KEY_RIGHT_2,
  KEY_A_2, KEY_B_2, KEY_X_2, KEY_Y_2, KEY_L_2, KEY_R_2,
  KEY_SELECT_2, KEY_START_2);
 set_config_string(NULL, "snes_pad_2_keys", keymapbuf);
}

void no_adjust(void){}

SCREEN screenmodes[8]={
 { 8,320,200,320,200,GFX_VGA   ,no_adjust  },  // 320x200x256 VGA
 { 8,320,200,320,200,GFX_VGA   ,no_adjust  },  // 320x200x256 VGA FIT (SQUASH)
 { 8,320,240,320,240,GFX_MODEX ,no_adjust  },  // 320x240x256 MODE-X
 { 8,320,200,256,239,GFX_VGA   ,Set256x239 },  // 256x239*256 VGA
 {16,320,200,320,200,GFX_VESA2L,no_adjust  },  // 320x200x16b SVGA
 {16,320,240,320,240,GFX_VESA2L,no_adjust  },  // 320x240x16b SVGA
 {16,640,480,640,480,GFX_VESA2L,no_adjust  },  // 640x480x16b SVGA
 {16,640,480,640,480,GFX_VESA2L,no_adjust  }   // 640x480x16b SVGA FIT (STRETCH)
};

// Used to configure the screen and set internal variables for rendering!
void SetGUIScreen(int ScreenMode){
 PC_SCREEN_MODE = ScreenMode;   // Save global reminder of new screen mode
 if (Allegro_Bitmap)    // If bitmap exists destroy it
  destroy_bitmap(Allegro_Bitmap);

 if (screenmodes[ScreenMode].set()){
  switch(ScreenMode){
   case 0:
   case 1:
   case 3:
    break;
   default:
   screenmodes[PC_SCREEN_MODE = 0].set();
  }
 }

 ScreenX = screenmodes[PC_SCREEN_MODE].w;
 ScreenY = screenmodes[PC_SCREEN_MODE].h;

 Allegro_Bitmap = create_bitmap(ScreenX, ScreenY);
 clear(Allegro_Bitmap);
 SNES_Screen = (unsigned char *)Allegro_Bitmap->dat;
 Reload_Bitmaps();  // We reload to convert to correct bit depth
}

void SetFullScreen(){
/*case 0:   // 320x200x256 VGA
  case 1:   // 320x200x256 VGA FIT (SQUASH)
  case 2:   // 320x240x256 MODE-X
  case 3:   // 256x239x256 VGA
  case 4:   // 320x200x16b SVGA
  case 5:   // 320x240x16b SVGA
  case 6:   // 640x480x16b SVGA
  case 7:   // 640x480x16b SVGA FIT (STRETCH) */
}

void cmdhelp(){
 const char syntax[] =
  "Usage: SNEeSe [switches] [romname8.ex3] [switches]\n"
  "switches:\n"
  " -fl  Force LoROM memory map\n"
  " -fh  Force HiROM memory map\n"
  " -fi  Force interleaved ROM (only supported for HiROM)\n"
  " -fn  Force non-interleaved ROM\n"
  " -fvn Force NTSC video standard\n"
  " -fvp Force PAL video standard\n"
  " -h   Copier header exists\n"
  " -n   No copier header\n";

 cout << syntax;
}

int main(int argc, char **argv){
 setvbuf(stdout, 0, _IONBF, 0);

#ifdef DEBUG
 LOCK_VARIABLE(Timer_Counter_Profile);
 LOCK_FUNCTION(Timer_Handler_Profile);
 LOCK_VARIABLE(Timer_Counter_FPS);
 LOCK_FUNCTION(Timer_Handler_FPS);
#endif

 LOCK_VARIABLE(Timer_Counter_Throttle);
 LOCK_FUNCTION(Timer_Handler_Throttle);

 cout << "SNEeSe, version " << SNEESE_VERSION << endl
      << "Release Date - " << RELEASE_DATE << endl
      << "Project founder: Lee Hammerton (Savoury SnaX)" << endl
      << "Original development: Lee Hammerton,Santeri Saarimaa,Charles Bilyue'" << endl
      << "Original sound system: Lee Hammerton,Brad Martin,archeide" << endl
      << "New development: Charles Bilyue'" << endl
      << "New sound system: Brad Martin" << endl
      << allegro_id << endl << endl;
 cout.flush();

 check_cpu();
 if (cpu_family < 4){
  cout << "SNEeSe requires a 486 or better!" << endl;
  return 1;
 }

 fname t;
 char *tv;
 int tc;
 int numnames;
 char *name[1];

 numnames = 0;
// if (argc == 1) cmdhelp();
 for (tc = 1; tc < argc; tc++){
  tv = argv[tc];
  if ((*tv == '-') || (*tv == '/')){
// if (++num_switches == argc) cmdhelp();
   switch (*(tv + 1)){
    case 'f':
    case 'F':
     switch (strlen(tv + 1)){
      case 2:
       switch(*(tv + 2)){
        case 'l': case 'L':  // Force LoROM
         ROM_memory_map = HiROM; break;
        case 'h': case 'H':  // Force HiROM
         ROM_memory_map = LoROM; break;
        case 'n': case 'N':  // Force non-interleaved
         ROM_interleaved = Off; break;
        case 'i': case 'I':  // Force interleaved
         ROM_interleaved = On; break;
        default: cout << "Invalid switch: " << tv << endl; cmdhelp(); exit(1);
       }
       break;
      case 3:
       switch(*(tv + 2)){
        case 'v': case 'V':  // Force video standard
         switch(*(tv + 3)){
          case 'n': case 'N':    // Force NTSC video standard
           ROM_video_standard = NTSC_video; break;
          case 'p': case 'P':    // Force PAL video standard
           ROM_video_standard = PAL_video; break;
          default: cout << "Invalid switch: " << tv << endl; cmdhelp(); exit(1);
         }
         break;
        default: cout << "Invalid switch: " << tv << endl; cmdhelp(); exit(1);
       }
       break;
      default: cout << "Invalid switch: " << tv << endl; cmdhelp(); exit(1);
     }
     break;
    case 'n': case 'N': // Force no header
     if (strlen(tv + 1) != 1){
      cout << "Invalid switch: " << tv << endl; cmdhelp(); exit(1);
     }
     ROM_has_header = Off; break;
    case 'h': case 'H': // Force header
     if (strlen(tv + 1) != 1){
      cout << "Invalid switch: " << tv << endl; cmdhelp(); exit(1);
     }
     ROM_has_header = On; break;
    case '?':
     if (strlen(tv + 1) != 1){
      cout << "Invalid switch: " << tv << endl;
     }
     cmdhelp(); exit(1);
    default:
     cout << "Invalid switch: " << tv << endl; cmdhelp(); exit(1);
   }
  } else {
   if (numnames < 1) name[numnames++] = tv;
   else { cmdhelp(); exit(1); }
  }
 }

 allegro_init();
 install_timer();
 int memoryneeded = Setup_Memory();
 if (memoryneeded){
  cout << "Out of memory. Need up to " << memoryneeded
       << " more bytes free!" << endl;
  return 1;
 }

 set_config_file("SNEeSe.CFG");	// Yup, config files exist

 // do before load config to ensure mouse ok.
 MOUSE_PLUGGED_IN = install_mouse() ? 1 : 0;

 LoadConfig();  // Sets up options to state based on the config file

 char *errormsg = GUI_init();
 if (errormsg){
  cout << errormsg;
  return -1;
 }

 if (numnames == 1){
  cout << "Attempting to load " << name[0] << endl;

  if (!LoadROM(name[0])){
   cout << "Failed to load cartridge ROM: " << name[0] << endl;
   return -1;
  }

  cout << endl << "Press any key to continue...";
  cout.flush();

  getch();
 }

 switch(SOUND_ENABLED){
  case 2:
   if (Install_Sound(1)) break;
   SOUND_ENABLED = 1;
  case 1:
   if (!Install_Sound(0)) SOUND_ENABLED = 0;
 }

 SetupTables();

// Take control of keyboard here (else getchar() won't bloody work!)
 install_keyboard();

// Warning... from now on the keyboard is not accessible except via 128 byte buffer!
 joy_type = JOY_TYPE_6BUTTON;
// initialise_joystick();

 SetGUIScreen(PC_SCREEN_MODE);   // PC_SCREEN_MODE set by config file

 atexit(&Wangle);
 for (;;){
  if (ROM_Loaded){
#define VOLUME_MONO 255 //170 - Butcha - prevent from muting mono mode
#define VOLUME_STEREO 255
   switch(SOUND_ENABLED){
    case 2: set_volume(VOLUME_MONO, -1); break;
    case 1: set_volume(VOLUME_STEREO, -1); break;
   }

   SetFullScreen();

   if (SOUND_ENABLED)
    sound_resume();

#ifdef DEBUG
   install_int_ex(Timer_Handler_Profile, MSEC_TO_TIMER(5));
   install_int_ex(Timer_Handler_FPS, SECS_TO_TIMER(1));
#endif
   install_int_ex(Timer_Handler_Throttle, Frametime);

   if (SPC_ENABLED) Make_SPC(); else Make_APU_Skipper();
   Do_CPU();    // This is the emulation core.

#ifdef DEBUG
   remove_int(Timer_Handler_Profile);
   remove_int(Timer_Handler_FPS);
#endif
   remove_int(Timer_Handler_Throttle);

   if (SOUND_ENABLED)
    sound_pause();
//  set_volume(0, -1);           // Ensures silent GUI
  }

  SetGUIScreen(PC_SCREEN_MODE); // PC_SCREEN_MODE set by config file
  if (!GUI()) break;    // Exit was pushed
 }

 set_gfx_mode(GFX_TEXT, 0, 0, 0, 0);

 remove_keyboard();
// Keyboard restored from now on... 

#ifdef DEBUG
 if (ROM_Loaded){
  DisplayStatus();

  if (Timer_Counter_Profile){
   cout << ((CPU_Profile_Test * 100.0) / Timer_Counter_Profile)
        << "% time in CPU cores" << endl;
  }
 }

 if (Timer_Counter_FPS){
  cout << "Frames Per Second - "
       << (Frames/Timer_Counter_FPS) << ".";
  if ((((Frames * 100) / Timer_Counter_FPS) % 100) < 10) cout << "0";
  cout << (((Frames*100)/Timer_Counter_FPS) % 100) << endl;
 }

// cout << "\nThanks for testing SNEeSe - look out for future releases.\n";
 cout << "\nThanks for using SNEeSe - look out for future releases.\n";
#else
 cout << "\nThanks for using SNEeSe - look out for future releases.\n";
#endif

 cout.flush();

 SaveSRAM(SRAMFilename);       // Save the Save RAM to file
 SaveConfig();

 return 0;
}
