/*  ROM IMAGE LOADING MECHANISM
    Original code by Savoury SnaX & Santeri Saarimaa
    New development by Charles Bilyue'

    Supports (with or without header):
     Single-part non-interleaved ROMs (LoROM and HiROM)
     Single-part interleaved HiROMs

    If a header exists, it is ignored

       Single-part ROM extensions
        SFC = Super FamiCom
        FIG = Pro Fighter
        SMC = Super MagiCom
        SWC = Super WildCard

        BIN = Binary

       Multi-part ROM extensions
        SF*a.058,SF*a.078 = Game Doctor
        *.1 = Other miscellaneous multi-part ROM

    TO DO:
        Multi-part ROM loading
        Interleaved LoROM loading (??? need info)
        Other memory maps
*/

#include <fstream.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "wrapaleg.h"   // Wrapper header for Allegro

#include "romload.h"
#include "helper.h"
#include "cpu.h"
#include "spc.h"
#include "mem.h"

#ifndef TRUE
#define TRUE         -1
#define FALSE        0
#endif

char *rom_romfile=0;
char *rom_romhilo=0;
char *rom_romname=0;
char *rom_romtype=0;
char *rom_romsize=0;
char *rom_sram=0;
char *rom_country=0;

extern void Timer_Handler_Throttle(...);
unsigned char *RomAddress;      // Address of SNES ROM

// Used to determine size of file for saving/loading, and to restrict writes
//  to non-existant SRAM
unsigned long SaveRamLength=0;

int Allocate_ROM();
unsigned NumBanks;
int ROM_format;
bool ROM_Loaded=FALSE;

unsigned long Frametime;

SNESRomInfoStruct RomInfoLo,RomInfoHi;

int LoadSRAM(char *SRAMFilename){
 fstream InFile(SRAMFilename,ios::in|ios::binary);

 if(!InFile) return 0;  // File aint there m8

 InFile.seekg(0,ios::beg);
 InFile.read(SRAM,SaveRamLength);   // Read in Save Ram

 return -1;
}

int SaveSRAM(char *SRAMFilename){
 if(!ROM_Loaded || !SRAMFilename || !SaveRamLength) return 0;

 fstream OutFile(SRAMFilename,ios::out|ios::binary);

 if(!OutFile) return 0; // File aint there m8

 OutFile.seekg(0,ios::beg);
 OutFile.write(SRAM,SaveRamLength); // Write out Save Ram

 return -1;
}

char *SRAMFilename=0;

char SaveExtension[4];

void CreateSRAMFilename(char *ROMFilename){
 int Count,Position;

 if(SRAMFilename) free(SRAMFilename);
 SRAMFilename=(char *)malloc(strlen(ROMFilename)+4);   // Get length + 4 for extension (.RAM is default)
 if(SRAMFilename){  // If unable to determine name, we skip save RAM!
  Count=strlen(ROMFilename)-1;
  while(Count>0 && ROMFilename[Count]!='\\') Count--;
  for(Position=0;Position<Count;Position++)
   SRAMFilename[Position]=ROMFilename[Position];
  while (ROMFilename[Count]!=0 && ROMFilename[Count]!='.'){
   SRAMFilename[Count]=ROMFilename[Count]; // Copy the rest of filename (& path)
   Count++;
  }
  SRAMFilename[Count+0]='.';
  SRAMFilename[Count+1]=SaveExtension[0];
  SRAMFilename[Count+2]=SaveExtension[1];
  SRAMFilename[Count+3]=SaveExtension[2];
  SRAMFilename[Count+4]=SaveExtension[3];
 }
}

void DisplayRomStats(SNESRomInfoStruct *RomInfo);

#ifdef DEBUG
#ifdef TRACKERS
unsigned char *HDMAAddress=0;
unsigned long LastHDMA;
unsigned char *InsAddress=0;
unsigned long LastIns;
#endif
#endif

// Allocates some memory used by emulation: tile caches,virtual output screen
int Setup_Memory(){
// - Align       = 1
// - 2bpl tiles  = 4
// - 4bpl tiles  = 2
// - 8bpl tiles  = 1
//         Total = 8
// - BKG         = (256+128)*240

 static unsigned char *AllocRAMAddress=0;
 unsigned char *temp;

#ifdef DEBUG
#ifdef TRACKERS
 HDMAAddress=(unsigned char *)malloc(1024);
 InsAddress=(unsigned char *)malloc(TRACKERS);
#endif
#endif

 if(AllocRAMAddress) free(AllocRAMAddress);

 AllocRAMAddress=(unsigned char *)malloc(((8)*65536)+(384*240));
 if(!AllocRAMAddress) return ((8)*65536)+(384*240);

 temp=  // Force 64k alignment
  (unsigned char *)(((unsigned long)(AllocRAMAddress+0xFFFF))&~0xFFFF);

 TileCache2=temp;
 temp+=0x40000;
 TileCache4=temp;
 temp+=0x20000;
 TileCache8=temp;
 temp+=0x10000;

// Background allocated such that there is space for 8 tiles on either side of
// the screen, this allows the rendering to function without breaking up any
// of the plotters!

 BKG=temp;
 temp+=(256+128)*240;

 return 0;
}

// NumBanks is 64k blocks
int Allocate_ROM(){
 // ROM is now dynamically allocated. This is to add large ROM support
 // without raising RAM requirements when using smaller ROMs.
 static unsigned char *AllocROMAddress=0;

 // De-allocate any previous memory
 if(AllocROMAddress) free(AllocROMAddress);

 AllocROMAddress=(unsigned char *)malloc((NumBanks+1)*65536);
 if(!AllocROMAddress) return (NumBanks+1)*65536;

 RomAddress=  // Force 64k alignment
  (unsigned char *)(((unsigned long)(AllocROMAddress+0xFFFF))&~0xFFFF);
 return 0;
}

// Resets memory heaps
void Reset_Memory(){
#ifdef DEBUG
#ifdef TRACKERS
 // Reset HDMA tracking buffer
 memset(HDMAAddress,0xFF,1024); LastHDMA=0;
 // Reset CPU instruction tracking buffer
 memset(InsAddress,0xFF,TRACKERS); LastIns=0;
#endif
#endif

 // Reset SPC address space to 0xFF
 memset(SPCRAM,0xFF,65536);

 // Reset WRAM to 0x55
 memset(WRAM,0x55,131072);

 // Reset VRAM to 0x55
 memset(VRAM,0x55,65536);

 // Reset blank ROM space to 0xFF
 memset(Blank,0xFF,65536);

 // Clear virtual SNES output screen
 memset(BKG,0,384*240);

 // Clear OAM
 memset(OAM,0,512+32);
}

// Resets Save RAM
void Reset_SRAM(){
 // Reset SRAM to 0xAA
 memset(SRAM,0xAA,131072);
}

unsigned long SRAM_Mask;

void Set_LoROM_Map(void){
 memcpy(Read_BankMapping,LoROM_Blank,256*4);
 memcpy(Write_BankMapping,LoROM_Write,256*4);

 SRAM_Mask=SaveRamLength-1;

 unsigned a;

 if(SaveRamLength){
  for(int i=0x70;i<0x74;i++){
   Read_BankMapping[i]=&SRAM_READ;
   Write_BankMapping[i]=&SRAM_WRITE;
  }
 } else {
  for(int i=0x70;i<0x74;i++){
   Read_BankMapping[i]=&UNSUPPORTED_READ;
   Write_BankMapping[i]=&UNSUPPORTED_WRITE;
  }
 }

 RomReadAdjust   =RomAddress;
 RomReadAdjust2  =RomAddress-(NumBanks<<16)-0x8000;
 RomHiReadAdjust =RomAddress-0x800000;
 RomHiReadAdjust2=RomAddress-(NumBanks<<16)-0x808000;

 for (a=0;a<0x100*2;a++) Read_BankOffset[a]=(void *) Blank;
 for(a=0;a<0x40;a++){
  Read_BankOffset[2*a       ]=(void *) WRAM;
  Read_BankOffset[2*(a+0x80)]=(void *) WRAM;
 }
 for (a=0;a<NumBanks;a++){
  Read_BankMapping[a]=&READ_00_3F_No_SRAM;
  Read_BankMapping[a+0x80]=&READ_80_BF_No_SRAM;
  Read_BankOffset[1+2*a       ]=(void *) (RomAddress+(a<<16));
  Read_BankOffset[1+2*(a+0x80)]=(void *) (RomAddress+(a<<16));
 }

 for (a=NumBanks;a<NumBanks*2;a++){
  if(a<64){
   Read_BankMapping[a]=&READ_00_3F_2;
   Read_BankMapping[a+0x80]=&READ_80_BF_2;
   Read_BankOffset[1+2*a       ]=(void *) (RomAddress+((a-NumBanks)<<16)-0x8000);
   Read_BankOffset[1+2*(a+0x80)]=(void *) (RomAddress+((a-NumBanks)<<16)-0x8000);
  } else {
   if(a<96){
    Read_BankMapping[a]=&READ_40_5F_2;
    Read_BankOffset[a*2+1]=(void *) (RomAddress+((a-NumBanks)<<16)-0x8000);
   }
   Read_BankMapping[a+0x80]=&READ_C0_FF_2;
   Read_BankOffset[(a+0x80)*2+1]=(void *) (RomAddress+((a-NumBanks)<<16)-0x8000);
  }
 }
 Read_BankOffset[0x7E*2  ]=(void *) WRAM;
 Read_BankOffset[0x7E*2+1]=(void *) WRAM;
 Read_BankOffset[0x7F*2  ]=(void *) (WRAM+0x10000);
 Read_BankOffset[0x7F*2+1]=(void *) (WRAM+0x10000);
}

void Set_HiROM_Map(void){
 memcpy(Read_BankMapping,HiROM_Blank,256*4);
 memcpy(Write_BankMapping,HiROM_Write,256*4);

 SRAM_Mask=SaveRamLength>=0x2000 ? 0x1FFF : SaveRamLength-1;

 unsigned a;

 RomReadAdjust      =RomAddress;
 HRomLinReadAdjust  =RomAddress-0x400000;
 RomHiReadAdjust    =RomAddress-0x800000;
 HRomHiLinReadAdjust=RomAddress-0xC00000;

 for(a=0;a<NumBanks;a++){  // Setup mapper for ROM banks
  if(a<0x20) Read_BankMapping[a+0x40]=HiROM_Read[a+0x40];
  if(a<0x30){
   Read_BankMapping[a]=HiROM_Read[a];
   Read_BankMapping[a+0x80]=HiROM_Read[a+0x80];
  }
  Read_BankMapping[a+0xC0]=HiROM_Read[a+0xC0];
 }

 switch(SaveRamLength){ // Setup mapper for rest of 32k ROM banks + SRAM
  case 0:       // No SRAM
   for(a=0x30;a<NumBanks;a++){
    Read_BankMapping [a]     =HiROM_Read_No_SRAM_ROM_30;
    Read_BankMapping [a+0x80]=HiROM_Read_No_SRAM_ROM_B0;
    Write_BankMapping[a]     =HiROM_Write_No_SRAM_30;
    Write_BankMapping[a+0x80]=HiROM_Write_No_SRAM_B0;
   }
   for(a=NumBanks;a<0x40;a++){
    Read_BankMapping [a]     =HiROM_Read_No_SRAM_Blank_30;
    Read_BankMapping [a+0x80]=HiROM_Read_No_SRAM_Blank_B0;
    Write_BankMapping[a]     =HiROM_Write_No_SRAM_30;
    Write_BankMapping[a+0x80]=HiROM_Write_No_SRAM_B0;
   }
   break;
  case 0x800:   // 2k SRAM
  case 0x1000:  // 4k SRAM
  case 0x2000:  // 8k SRAM
   for(a=0x30;a<NumBanks;a++){
    Read_BankMapping [a]     =HiROM_Read_SRAM_ROM_30[0];
    Read_BankMapping [a+0x80]=HiROM_Read_SRAM_ROM_B0[0];
    Write_BankMapping[a]     =HiROM_Write_SRAM_30[0];
    Write_BankMapping[a+0x80]=HiROM_Write_SRAM_B0[0];
   }
   for(a=NumBanks;a<0x40;a++){
    Read_BankMapping [a]     =HiROM_Read_SRAM_Blank_30[0];
    Read_BankMapping [a+0x80]=HiROM_Read_SRAM_Blank_B0[0];
    Write_BankMapping[a]     =HiROM_Write_SRAM_30[0];
    Write_BankMapping[a+0x80]=HiROM_Write_SRAM_B0[0];
   }
   break;
  case 0x4000:  // 16k SRAM
   for(a=0x30;a<NumBanks;a++){
    Read_BankMapping [a]     =HiROM_Read_SRAM_ROM_30[a&1];
    Read_BankMapping [a+0x80]=HiROM_Read_SRAM_ROM_B0[a&1];
    Write_BankMapping[a]     =HiROM_Write_SRAM_30[a&1];
    Write_BankMapping[a+0x80]=HiROM_Write_SRAM_B0[a&1];
   }
   for(a=NumBanks;a<0x40;a++){
    Read_BankMapping [a]     =HiROM_Read_SRAM_Blank_30[a&1];
    Read_BankMapping [a+0x80]=HiROM_Read_SRAM_Blank_B0[a&1];
    Write_BankMapping[a]     =HiROM_Write_SRAM_30[a&1];
    Write_BankMapping[a+0x80]=HiROM_Write_SRAM_B0[a&1];
   }
   break;
  case 0x8000:  // 32k SRAM
   for(a=0x30;a<NumBanks;a++){
    Read_BankMapping [a]     =HiROM_Read_SRAM_ROM_30[a&3];
    Read_BankMapping [a+0x80]=HiROM_Read_SRAM_ROM_B0[a&3];
    Write_BankMapping[a]     =HiROM_Write_SRAM_30[a&3];
    Write_BankMapping[a+0x80]=HiROM_Write_SRAM_B0[a&3];
   }
   for(a=NumBanks;a<0x40;a++){
    Read_BankMapping [a]     =HiROM_Read_SRAM_Blank_30[a&3];
    Read_BankMapping [a+0x80]=HiROM_Read_SRAM_Blank_B0[a&3];
    Write_BankMapping[a]     =HiROM_Write_SRAM_30[a&3];
    Write_BankMapping[a+0x80]=HiROM_Write_SRAM_B0[a&3];
   }
   break;
  case 0x10000: // 64k SRAM
   for(a=0x30;a<NumBanks;a++){
    Read_BankMapping [a]     =HiROM_Read_SRAM_ROM_30[a&7];
    Read_BankMapping [a+0x80]=HiROM_Read_SRAM_ROM_B0[a&7];
    Write_BankMapping[a]     =HiROM_Write_SRAM_30[a&7];
    Write_BankMapping[a+0x80]=HiROM_Write_SRAM_B0[a&7];
   }
   for(a=NumBanks;a<0x40;a++){
    Read_BankMapping [a]     =HiROM_Read_SRAM_Blank_30[a&7];
    Read_BankMapping [a+0x80]=HiROM_Read_SRAM_Blank_B0[a&7];
    Write_BankMapping[a]     =HiROM_Write_SRAM_30[a&7];
    Write_BankMapping[a+0x80]=HiROM_Write_SRAM_B0[a&7];
   }
   break;
 }

 for (a=0;a<0x100*2;a++) Read_BankOffset[a]=(void *) Blank;
 for(a=0;a<0x40;a++){
  Read_BankOffset[2*a       ]=(void *) WRAM;
  Read_BankOffset[2*(a+0x80)]=(void *) WRAM;
 }
 for (a=0;a<NumBanks;a++){
  // Set 32k banks
  Read_BankOffset[2*a       +1]=(void *) (RomAddress+(a<<16));
  Read_BankOffset[2*(a+0x80)+1]=(void *) (RomAddress+(a<<16));
  // Set 64k banks
  Read_BankOffset[2*(a+0xC0)  ]=(void *) (RomAddress+(a<<16));
  Read_BankOffset[2*(a+0xC0)+1]=(void *) (RomAddress+(a<<16));
  if(a<32){
   Read_BankOffset[2*(a+0x40)  ]=(void *) (RomAddress+(a<<16));
   Read_BankOffset[2*(a+0x40)+1]=(void *) (RomAddress+(a<<16));
  }
 }
 Read_BankOffset[0x7E*2  ]=(void *) WRAM;
 Read_BankOffset[0x7E*2+1]=(void *) WRAM;
 Read_BankOffset[0x7F*2  ]=(void *) (WRAM+0x10000);
 Read_BankOffset[0x7F*2+1]=(void *) (WRAM+0x10000);
}

char *ROMFilename=0;

int ROM_has_header = Undetected;
/* Off = LoROM, On = HiROM, Undetected = autodetect */
int ROM_memory_map = Undetected;
int ROM_interleaved = Undetected;
int ROM_video_standard = Undetected;

int LoadROM(char *Filename){
 SaveSRAM(SRAMFilename);   // Ensures SRAM saved before loading new ROM
 ROM_Loaded=FALSE;
 SaveRamLength=0;
 if(SRAMFilename){ free(SRAMFilename); SRAMFilename=0; }
 if(ROMFilename){ free(ROMFilename); ROMFilename=0; }

 ROMFilename=(char *) malloc(strlen(Filename));
 if(ROMFilename) strcpy(ROMFilename,Filename);
 if(!rom_romfile) rom_romfile=(char *)malloc(100);
 if(!rom_romhilo) rom_romhilo=(char *)malloc(100);
 if(!rom_romtype) rom_romtype=(char *)malloc(50);
 if(!rom_romsize) rom_romsize=(char *)malloc(50);
 if(!rom_romname) rom_romname=(char *)malloc(50);
 if(!rom_sram)    rom_sram   =(char *)malloc(50);
 if(!rom_country) rom_country=(char *)malloc(50);

 strcpy(rom_romfile,"File name: ");
 strcat(rom_romfile,Filename);
 rom_romfile[33]=0;

 FILE *infile=fopen(Filename,"rb");
 if(!infile) return FALSE;  // File aint there m8

 unsigned ROM_start;        // This is where the ROM code itself starts.

 fseek(infile, 0, SEEK_END);
 int infilesize = ftell(infile);
 // Easiest way to detect ROM header
  // Improvement suggested by Gridle (implemented 22/4/99)
  // check against even multiple of 1k instead of 32k
 // If ROM filesize is even multiple of 1k, assume no header
 // If ROM filesize is even multiple of 1k, +512, assume header
 if ((((infilesize % 1024) == 0) || (ROM_has_header == Off)) &&
  (ROM_has_header != On)) ROM_start = 0;
 else if (((infilesize % 1024) == ROM_Header_Size) || (ROM_has_header == On))
  ROM_start = ROM_Header_Size;
 else { // Basic header detection failure
/* Gridle 04/03/1998 (dmy)

ROM header in SMC files at 0x100 - 0x1FF usually is zero. I couldn't find
any other reliable way to check if the ROM is really SMC or BIN.

Originally I checked everything from 0x10 - 0x1FF, but one of my ROMs
(Brainies) had some stupid pirate group stuff just before 0x100, so I
changed this to search from 0x100 - 0x1FF.

This is now changed to 0x100 - 0x1F9 because of the king arthur ROM.

Charles Bilyue' 06/01/1999 (dmy)

(Check was later changed to 0x110 - 0x1F9 for some reason?)

LoROM checksums at 7FDC/7FDE, 81DC/81DE with header
HiROM checksums at FFDC/FFDE, 101DC/101DE with header

*/
  ROM_Header Header;     // For checking if the ROM has a header

  fseek(infile,0,SEEK_SET);
  fread(&Header,ROM_Header_Size,1,infile);

  ROM_start=ROM_Header_Size;

  for (unsigned cnt=0x110-11;cnt<0x1FA-11;cnt++)
   if(Header.RestOfHeader[cnt]!=0){ ROM_start=0; break; }
 }
 
 if(ROM_start == ROM_Header_Size)
  printf("Header detected and ignored.\n");
 else
  printf("No header detected.\n");

 ROM_format = Undetected;

 fseek(infile,0x7FC0+ROM_start,SEEK_SET);
 fread(&RomInfoLo,sizeof(SNESRomInfoStruct),1,infile);
 fseek(infile,0xFFC0+ROM_start,SEEK_SET);
 fread(&RomInfoHi,sizeof(SNESRomInfoStruct),1,infile);

 if((RomInfoLo.Checksum^RomInfoLo.Complement)==0xFFFF){
  if((RomInfoLo.ROM_makeup&0x0F)==1){
   strcpy(rom_romhilo,"Interleaved HiROM detected");
   ROM_format = HiROM_Interleaved;
  } else {
   strcpy(rom_romhilo,"LoROM detected");
   ROM_format = LoROM;
  }
 } else {
  if(((RomInfoHi.Checksum^RomInfoHi.Complement)==0xFFFF)){
   strcpy(rom_romhilo,"HiROM detected");
   ROM_format = HiROM;
  } else {
   strcpy(rom_romhilo,"Detection failed, using LoROM");
   ROM_format = LoROM;
  }
 }

 switch (ROM_interleaved){
  case On: ROM_format |= Interleaved; break;
  case Off: ROM_format &= ~Interleaved; break;
 }

 switch (ROM_memory_map){
  case HiROM:
   ROM_format |= HiROM; strcpy(rom_romhilo,"HiROM forced"); break;
  case LoROM:
   ROM_format &= ~HiROM; strcpy(rom_romhilo,"LoROM forced"); break;
 }

 NumBanks=(((infilesize-ROM_start)+65535)/65536);
 if(NumBanks>64) NumBanks=64;  // Maximum 32Mbit ROM size for LoROM
 if(Allocate_ROM())    // Dynamic allocation of ROM
  return FALSE;        // return false if no memory left

 fseek(infile,ROM_start,SEEK_SET);

 switch(ROM_format){
  case HiROM:
   DisplayRomStats(&RomInfoHi);
   Set_HiROM_Map();

   for (unsigned cnt=0;cnt<NumBanks;cnt++){  // Read in ROM
    unsigned readcnt=fread(RomAddress+cnt*65536,1,65536,infile);
    if(readcnt<65536) memset(RomAddress+cnt*65536+readcnt,0xFF,65536-readcnt);
   }
   break;
  case HiROM_Interleaved:
   DisplayRomStats(&RomInfoLo);
   Set_HiROM_Map();

   // Read in as 32k blocks and deinterleave
   for (unsigned cnt=0;cnt<NumBanks;cnt++){  // Read first half
    unsigned readcnt=fread(RomAddress+cnt*65536+32768,1,32768,infile);
    if(readcnt<32768) memset(RomAddress+cnt*65536+32768+readcnt,0xFF,32768-readcnt);
   }

   for (unsigned cnt=0;cnt<NumBanks;cnt++){  // Read second half
    unsigned readcnt=fread(RomAddress+cnt*65536,1,32768,infile);
    if(readcnt<32768) memset(RomAddress+cnt*65536+readcnt,0xFF,32768-readcnt);
   }
   break;
  case LoROM_Interleaved:
   printf("Interleaved LoROM not supported - basic LoROM loader used\n");
  case LoROM:
  default:
   DisplayRomStats(&RomInfoLo);
   Set_LoROM_Map();

   // Read in as 32k blocks and interleave
   for (unsigned cnt=0;cnt<NumBanks;cnt++){  // Read first half
    unsigned readcnt=fread(RomAddress+cnt*65536+32768,1,32768,infile);
    if(readcnt<32768) memset(RomAddress+cnt*65536+32768+readcnt,0xFF,32768-readcnt);
   }

   for (unsigned cnt=0;cnt<NumBanks;cnt++){  // Read second half
    unsigned readcnt=fread(RomAddress+cnt*65536,1,32768,infile);
    if(readcnt<32768) memset(RomAddress+cnt*65536+readcnt,0xFF,32768-readcnt);
   }
   break;
 }
 fclose(infile);

 ROM_Loaded=TRUE;
 Reset_Memory();
 Reset_SRAM();
 install_int_ex(Timer_Handler_Throttle,Frametime);
 Reset_CPU();
 remove_int(Timer_Handler_Throttle);
 Reset_SPC();

 CreateSRAMFilename(Filename);
 if(SRAMFilename && SaveRamLength)
  LoadSRAM(SRAMFilename);     // This loads in the Save RAM

 printf("\nRES: %04X\nNMI: %04X,%04X\nIRQ: %04X,%04X\n",
  _65816_PC,NMI_Nvector,NMI_Evector,IRQ_Nvector,IRQ_Evector);
 printf("\nBRK: %04X,%04X\nCOP: %04X,%04X",
  BRK_Nvector,IRQ_Evector,COP_Nvector,COP_Evector);
 return TRUE;
}

char *TypeTable[]={
 "ROM",
 "ROM+RAM",
 "ROM+SRAM",
 "ROM+DSP1",
 "ROM+RAM+DSP1",
 "ROM+SRAM+DSP1",
 "Unknown",
 "Unknown FX chip",
 "FX-V1",
 "FX(Argonaut)",
 "FX-V2",
 "FX(SA-1)",
 "FX(?KSS?)",
 "FX(Capcom)",
 "FX(S-DD1)",
 "FX(PLGS)",
 "FX(Gameboy)",
 "FX(BS-X)",
 "FX(Capcom C4)",
 "FX(SETA) (DSP2?)"
};

char *CountryTable[]={
 "Japan",
 "USA",
 "Europe, Oceania, Asia",
 "Sweden",
 "Finland",
 "Denmark",
 "France",
 "Holland",
 "Spain",
 "Germany, Austria, Switz.",
 "Italy",
 "Hong Kong, China",
 "Indonesia",
 "Korea"
};

extern "C" unsigned FPSMaxTicks;
void DisplayRomStats(SNESRomInfoStruct *RomInfo){
 char TempBuffer[30];

 // ROM Type

 strcpy(rom_romtype,"ROM Type: ");
 switch(RomInfo->ROM_type){
  case 0:   // ROM
  case 1:   // ROM/RAM
  case 2:   // ROM/SRAM
  case 3:   // ROM/DSP1
  case 4:   // ROM/RAM/DSP1
  case 5:   // ROM/SRAM/DSP1
   strcat(rom_romtype,TypeTable[RomInfo->ROM_type]); break;
  case 0x10 ... 0x12:   case 0x16:
  case 0x20 ... 0x26:   case 0x30 ... 0x34:
  case 0x40 ... 0x42:   case 0x44:  case 0x46:
  case 0x50 ... 0x54:   case 0x56:
  case 0x60 ... 0x66:   case 0x70 ... 0x76:
  case 0x80 ... 0x86:   case 0x90 ... 0x96:
  case 0xA0 ... 0xA6:   case 0xB0 ... 0xB6:
  case 0xC0 ... 0xC6:   case 0xD0 ... 0xD6:
  case 0xE0 ... 0xE2:   case 0xE4:  case 0xE6:
  case 0xF0 ... 0xF2:   case 0xF4 ... 0xF5:
            // Unknown FX chip
   strcat(rom_romtype,TypeTable[7]);
  case 0x13:    // FX-V1
   strcat(rom_romtype,TypeTable[8]); break;
  case 0x14:    // FX(Argonaut)
   strcat(rom_romtype,TypeTable[9]); break;
  case 0x15:    // FX-V2
   strcat(rom_romtype,TypeTable[10]); break;
  case 0x35:    // FX(SA-1)
   strcat(rom_romtype,TypeTable[11]); break;
  case 0x36:    // FX(?KSS?)
   strcat(rom_romtype,TypeTable[12]); break;
  case 0x43:    // FX(Capcom)
   strcat(rom_romtype,TypeTable[13]); break;
  case 0x45:    // FX(S-DD1)
   strcat(rom_romtype,TypeTable[14]); break;
  case 0x55:    // FX(PLGS)
   strcat(rom_romtype,TypeTable[15]); break;
  case 0xE3:    // FX(Gameboy)
   strcat(rom_romtype,TypeTable[16]); break;
  case 0xE5:    // FX(BS-X)
   strcat(rom_romtype,TypeTable[17]); break;
  case 0xF3:    // FX(Capcom C4)
   strcat(rom_romtype,TypeTable[18]); break;
  case 0xF6:    // FX(SETA) (DSP2?)
   strcat(rom_romtype,TypeTable[19]); break;
  default:      // Unknown
   strcat(rom_romtype,TypeTable[6]);
 }

 // ROM Size

 strcpy(rom_romsize,"ROM Size : ");
 strcat(rom_romsize,itoa(1 << ((RomInfo->ROM_size)-7),TempBuffer,10));
 strcat(rom_romsize,"Mbits");

 // ROM SRAM size

 SaveRamLength=RomInfo->SRAM_size;
 if(((RomInfo->ROM_type)==2) || ((RomInfo->ROM_type)>=5) &&
  SaveRamLength){
  SaveRamLength=1024 << SaveRamLength;

  strcpy(rom_sram,"SRAM Size: ");
  strcat(rom_sram,itoa(SaveRamLength*8/1024,TempBuffer,10));
  strcat(rom_sram,"Kbits");
  if(SaveRamLength>65536) SaveRamLength=65536;
 } else {
  SaveRamLength=0;
  strcpy(rom_sram,"SRAM Size: No SRAM");
 }

 // ROM Country

 IRQTrip = LineCycles = 340 * 4;    // 0.37

 if (((RomInfo->Country_code < 2) || (ROM_video_standard == NTSC_video))
 && (ROM_video_standard != PAL_video)){
  strcpy(rom_country, "Country: ");
  strcat(rom_country, CountryTable[ RomInfo->Country_code ]);
  strcat(rom_country,
   (ROM_video_standard != NTSC_video) ? " (NTSC)" : " (NTSC forced)");
  SNES_COUNTRY = 0x00;
  Event_Handler_Table = Event_Handler_Table_NTSC;

  Frametime = BPS_TO_TIMER(60);
  FPSMaxTicks = 60;
 } else if (((RomInfo->Country_code < 14) || (ROM_video_standard == PAL_video))
 && (ROM_video_standard != NTSC_video)){
  strcpy(rom_country, "Country: ");
  strcat(rom_country, CountryTable[ RomInfo->Country_code ]);
  strcat(rom_country,
   (ROM_video_standard != PAL_video) ? " (PAL)" : " (PAL forced)");
  SNES_COUNTRY = 0x10;
  Event_Handler_Table = Event_Handler_Table_PAL;

  Frametime = BPS_TO_TIMER(50);
  FPSMaxTicks = 50;
 } else {
  strcpy(rom_country, "Country: Unknown (Using NTSC)");
  SNES_COUNTRY = 0x00;
  Event_Handler_Table = Event_Handler_Table_NTSC;

  Frametime = BPS_TO_TIMER(60);
  FPSMaxTicks = 60;
 }

 // ROM Name

 for(int i = 0; i < 21; i++)
  TempBuffer[i] = RomInfo->ROM_Title[i];

 TempBuffer[21] = 0;

 strcpy(rom_romname,"ROM Title: ");
 strcat(rom_romname,TempBuffer);

 // Print information

 printf("\n\n%s\n",rom_romname);
 printf("%s\n",rom_romtype);
 if((RomInfo->ROM_type)>2)
  printf("Some extra hardware used by this ROM is not yet supported!\n");
 printf("%s\n",rom_romhilo);
 printf("%s\n",rom_romsize);
 printf("%s\n",rom_sram);
 printf("%s",rom_country);

 // Ugly things, but helps with 256x240 in ROM information window
 rom_romhilo[33]=0;
 rom_country[33]=0;
}

extern unsigned long SPC_Cycles asm("SPC_Cycles");
extern unsigned long SPC_last_cycles asm("SPC_last_cycles");
extern unsigned long SPC_CPU_cycles asm("SPC_CPU_cycles");

#include "sound.h"

void Wangle(void){
 if(ROM_Loaded){
#ifdef DEBUG
#ifdef TRACKERS
  fstream HDMA_DUMP("HDMA.DMP",ios::bin|ios::out);
  HDMA_DUMP.write(HDMAAddress+LastHDMA,1024-LastHDMA);
  HDMA_DUMP.write(HDMAAddress,LastHDMA);
  HDMA_DUMP.close();
  // This saves the CPU-tracker dump!
  fstream INS_DUMP("C:\\INS.DMP",ios::bin|ios::out);
  INS_DUMP.write(InsAddress+LastIns,(TRACKERS)-LastIns);
  INS_DUMP.write(InsAddress,LastIns);
  INS_DUMP.close();
#endif
  fstream SPC_DUMP("SPC.DMP",ios::bin|ios::out);
  SPC_DUMP.write(SPCRAM,65536);
  SPC_DUMP.close();

  fstream WRAM_DUMP("WRAM.DMP",ios::bin|ios::out);
  WRAM_DUMP.write(WRAM,131072);
  WRAM_DUMP.close();

  fstream VRAM_DUMP("VRAM.DMP",ios::bin|ios::out);
  VRAM_DUMP.write(VRAM,65536);
  VRAM_DUMP.close();
#endif
 }
}
